// $Id$
/**
* Copyright (C) 2007 EDIT
* European Distributed Institute of Taxonomy 
* http://www.e-taxonomy.eu
* 
* The contents of this file are subject to the Mozilla Public License Version 1.1
* See LICENSE.TXT at the top of this package for the full license terms.
*/

package eu.etaxonomy.taxeditor.annotatedlineeditor;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.source.Annotation;
import org.eclipse.jface.text.source.AnnotationModel;
import org.eclipse.jface.text.source.AnnotationModelEvent;

import eu.etaxonomy.taxeditor.model.MessagingUtils;

/**
 * An <code>AnnotationModel</code> which holds <code>LineAnnotation</code>'s.
 * <p>
 * Major difference with <code>AnnotationModel</code> is that <code>removeAnnotation(Annotation annotation,
 * boolean fireModelChanged)</code> adds annotations marked for removal to a collection for further
 * processing, i.e. when the document is saved. This collection is accessed via <code>getDeletedAnnotations()</code>
 * and <code>clearDeletedAnnotations()</code>.
 *
 * @author p.ciardelli
 * @created 25.06.2009
 * @version 1.0
 */
public class LineAnnotationModel extends AnnotationModel {
	
	private Comparator<Annotation> comparator;
	private IEntityCreator<?> entityCreator;
	private Set<LineAnnotation> deletedAnnotations = new HashSet<LineAnnotation>();

	private ILineDisplayStrategy lineDisplayStrategy;
	
	/**
	 * <p>Constructor for LineAnnotationModel.</p>
	 *
	 * @param lineDisplayStrategy a {@link eu.etaxonomy.taxeditor.annotatedlineeditor.ILineDisplayStrategy} object.
	 */
	public LineAnnotationModel(ILineDisplayStrategy lineDisplayStrategy) {
		super();
		this.lineDisplayStrategy = lineDisplayStrategy;
	}
	
	/**
	 * Changes the annotation's type and fires model changed
	 * notification. Resulting redraw will cause any change
	 * in annotation's decoration to be displayed.
	 *
	 * @param annotation a {@link org.eclipse.jface.text.source.Annotation} object.
	 * @param type a {@link java.lang.String} object.
	 */
	public void changeAnnotationType(Annotation annotation, String type) {
		annotation.setType(type);
		fireModelChanged(new AnnotationModelEvent(this));
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.jface.text.source.AnnotationModel#removeAnnotation(org.eclipse.jface.text.source.Annotation, boolean)
	 */
	/** {@inheritDoc} */
	protected void removeAnnotation(Annotation annotation,
			boolean fireModelChanged) {
		// Set data model element for deletion
		annotation.markDeleted(true);
		if (annotation instanceof LineAnnotation) {
			deletedAnnotations.add((LineAnnotation) annotation);
		}
		super.removeAnnotation(annotation, fireModelChanged);
	}
		
	/**
	 * <p>removeTypeFromAllAnnotations</p>
	 *
	 * @param type a {@link java.lang.String} object.
	 */
	public void removeTypeFromAllAnnotations(String type) {
		Iterator<?> iterator = getAnnotationIterator();
		while (iterator.hasNext()) {
			Annotation annotation = ((Annotation) iterator.next());
			if (annotation.getType().equals(type)) {
				annotation.setType(Annotation.TYPE_UNKNOWN);
			}
		}
	}
	
	/**
	 * <p>getAllAnnotationsOfType</p>
	 *
	 * @param type a {@link java.lang.String} object.
	 * @return a {@link java.util.Set} object.
	 */
	public Set<LineAnnotation> getAllAnnotationsOfType(String type) {
		Set<LineAnnotation> candidates = new HashSet<LineAnnotation>();
		Iterator<?> iterator = getAnnotationIterator();
		while (iterator.hasNext()) {
			LineAnnotation annotation = ((LineAnnotation) iterator.next());
			if (annotation.getType().equals(type)) {
				candidates.add(annotation);
			}
		}	
		return candidates;	
	}	
	
	/**
	 * <p>getFirstAnnotationOfType</p>
	 *
	 * @param type a {@link java.lang.String} object.
	 * @return a {@link org.eclipse.jface.text.source.Annotation} object.
	 */
	public Annotation getFirstAnnotationOfType(String type) {
		Iterator<?> iterator = getAnnotationIterator();
		while (iterator.hasNext()) {
			Annotation annotation = ((Annotation) iterator.next());
			if (annotation.getType().equals(type)) {
				return annotation;
			}
		}		
		return null;
	}
	
	/**
	 * Gets first LineAnnotation in line
	 *
	 * @param line a int.
	 * @param document a {@link org.eclipse.jface.text.IDocument} object.
	 * @return a {@link org.eclipse.jface.text.source.Annotation} object.
	 */
	public Annotation getAnnotationAtLine(int line, IDocument document) {
//		Annotation annotation = null;
		try {
			int offset = document.getLineOffset(line);
			int length = document.getLineLength(line);
						
			Iterator<?> iterator = getAnnotationIterator(offset, length, true, true);
			while (iterator.hasNext()) {
				Object next = iterator.next();
				if (next instanceof LineAnnotation) {
					return (LineAnnotation) next;
				}
//				annotation = (Annotation) iterator.next();
			}
		} catch (BadLocationException e) {
			// do nothing
		}
		return null;
	}
		
	/**
	 * <p>getOrderedAnnotations</p>
	 *
	 * @return a {@link java.util.List} object.
	 */
	public List<Annotation> getOrderedAnnotations() {
		List<Annotation> list = new ArrayList<Annotation>();
		Iterator<?> iterator = getAnnotationIterator();
		while (iterator.hasNext()) {
			list.add((Annotation) iterator.next());
		}
		Collections.sort(list, getAnnotationComparator());
		return list;
	}
		
	/**
	 * <p>printAnnotations</p>
	 */
	public void printAnnotations() {
		Logger logger = MessagingUtils.getLog4JLogger(getClass());
		logger.debug("------------------------");
		logger.debug("Active annotations");
		logger.debug("------------------------");
		List<Annotation> list = getOrderedAnnotations();
		
		for (Annotation annotation : list) {
			logger.debug(
					(annotation.isMarkedDeleted() ?  "DELETED " : "") +
					(((LineAnnotation) annotation).isMarkedAsMerged() ?  "MERGED " : "") +
					(((LineAnnotation) annotation).isDirty() ?  "DIRTY " : "") +
					(((LineAnnotation) annotation).isMarkedAsNew() ?  "NEW " : "") +
					annotation.getText() + ": o " +
					getPosition(annotation).getOffset() + ", l " + 
					getPosition(annotation).getLength());
		}
		logger.debug("------------------------");
		logger.debug("Deleted annotations");
		logger.debug("------------------------");
		for (LineAnnotation annotation : deletedAnnotations) {
			logger.debug(
					(annotation.isMarkedDeleted() ?  "DELETED " : "") +
					(((LineAnnotation) annotation).isMarkedAsMerged() ?  "MERGED " : "") +
					(((LineAnnotation) annotation).isDirty() ?  "DIRTY " : "") +
					(((LineAnnotation) annotation).isMarkedAsNew() ?  "NEW " : "") +
					annotation.getText() + ": o ");			
		}
	}
	
	/**
	 * <p>getUndeletedAnnotations</p>
	 *
	 * @param offset a int.
	 * @param length a int.
	 * @return a {@link java.util.List} object.
	 */
	public List<LineAnnotation> getUndeletedAnnotations(int offset, int length) {
		
		List<LineAnnotation> list = new ArrayList<LineAnnotation>();
		
		Iterator<?> iterator = getAnnotationIterator(offset, length, true, true);
		while (iterator.hasNext()) {
			Object next = iterator.next();
			if (next instanceof LineAnnotation && !((Annotation) next).isMarkedDeleted()) {
				list.add((LineAnnotation) next); 
			}
		}
		Collections.sort(list, getAnnotationComparator());
		return list;
	}
	
	/**
	 * <p>Getter for the field <code>deletedAnnotations</code>.</p>
	 *
	 * @return a {@link java.util.Set} object.
	 */
	public Set<LineAnnotation> getDeletedAnnotations() {
		return deletedAnnotations;
	}
	
	/**
	 * <p>clearDeletedAnnotations</p>
	 */
	public void clearDeletedAnnotations() {
		deletedAnnotations.clear();
	}
	
	private Comparator<Annotation> getAnnotationComparator() {
		if (comparator == null) {
			this.comparator = new AnnotationComparator();
		}
		return comparator;
	}
	
	class AnnotationComparator implements Comparator<Annotation> {

		/* (non-Javadoc)
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(Annotation annotation1, Annotation annotation2) {
			if (annotation1.isMarkedDeleted() && !annotation2.isMarkedDeleted()) {
				return 1;
			}
			if (!annotation1.isMarkedDeleted() && annotation2.isMarkedDeleted()) {
				return -1;
			}
			int offset1 = LineAnnotationModel.this.
								getPosition(annotation1).getOffset();
			int offset2 = LineAnnotationModel.this.
								getPosition(annotation2).getOffset();
			if (offset1 > offset2) {
				return 1;
			} else if (offset1 < offset2) {
				return -1;
			} else {
				return 0;
			}
		}	
	}
	
	/**
	 * The annotation model is assigned responsibility for creating new containers
	 * from blocks of text.
	 *
	 * @param entityCreator a {@link eu.etaxonomy.taxeditor.annotatedlineeditor.IEntityCreator} object.
	 */
	public void setEntityCreator(IEntityCreator<?> entityCreator) {
		this.entityCreator = entityCreator;
	}
	
	/**
	 * Creates an annotation without adding it to the model.
	 *
	 * This object must have its <code>cdmEntityCreator</code> set.
	 *
	 * @param text a {@link java.lang.String} object.
	 * @return a {@link org.eclipse.jface.text.source.Annotation} object.
	 */
	public Annotation createAnnotation(String text) {
		LineAnnotation annotation = null;
		if (entityCreator != null) {
			annotation = new LineAnnotation(entityCreator.createEntity(text), lineDisplayStrategy);
			annotation.markAsNew(true);
		}
		return annotation;
	}

	/**
	 * Creates an annotation without adding it to the model.
	 *
	 * @param entity a {@link java.lang.Object} object.
	 * @return a {@link org.eclipse.jface.text.source.Annotation} object.
	 */
	public Annotation createAnnotation(Object entity) {
		LineAnnotation annotation = new LineAnnotation(entity, lineDisplayStrategy);
		annotation.markAsNew(true);
		return annotation;
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.jface.text.source.AnnotationModel#addAnnotation(org.eclipse.jface.text.source.Annotation, org.eclipse.jface.text.Position)
	 */
	/** {@inheritDoc} */
	@Override
	public void addAnnotation(Annotation annotation, Position position) {
		super.addAnnotation(annotation, position);
	}
	
	/**
	 * <p>getAnnotation</p>
	 *
	 * @param entity a {@link java.lang.Object} object.
	 * @return a {@link org.eclipse.jface.text.source.Annotation} object.
	 */
	public Annotation getAnnotation(Object entity) {
		if (entity != null) {
			Iterator iterator = getAnnotationIterator();
			while (iterator.hasNext()) {
				LineAnnotation annotation = (LineAnnotation) iterator.next();
				if (entity.equals(annotation.getEntity())) {
					return annotation;
				}
			}
		}
		return null;
	}
}
